# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: src/nitrokey/trussed/_bootloader/nrf52_upload/dfu/dfu-cc.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n>src/nitrokey/trussed/_bootloader/nrf52_upload/dfu/dfu-cc.proto\x12\x03\x64\x66u"6\n\x04Hash\x12 \n\thash_type\x18\x01 \x02(\x0e\x32\r.dfu.HashType\x12\x0c\n\x04hash\x18\x02 \x02(\x0c"C\n\x0e\x42ootValidation\x12!\n\x04type\x18\x01 \x02(\x0e\x32\x13.dfu.ValidationType\x12\x0e\n\x06\x62ytes_\x18\x02 \x02(\x0c"\xf8\x01\n\x0bInitCommand\x12\x12\n\nfw_version\x18\x01 \x01(\r\x12\x12\n\nhw_version\x18\x02 \x01(\r\x12\x12\n\x06sd_req\x18\x03 \x03(\rB\x02\x10\x01\x12\x19\n\x04type\x18\x04 \x01(\x0e\x32\x0b.dfu.FwType\x12\x0f\n\x07sd_size\x18\x05 \x01(\r\x12\x0f\n\x07\x62l_size\x18\x06 \x01(\r\x12\x10\n\x08\x61pp_size\x18\x07 \x01(\r\x12\x17\n\x04hash\x18\x08 \x01(\x0b\x32\t.dfu.Hash\x12\x17\n\x08is_debug\x18\t \x01(\x08:\x05\x66\x61lse\x12,\n\x0f\x62oot_validation\x18\n \x03(\x0b\x32\x13.dfu.BootValidation"\x1f\n\x0cResetCommand\x12\x0f\n\x07timeout\x18\x01 \x02(\r"i\n\x07\x43ommand\x12\x1c\n\x07op_code\x18\x01 \x01(\x0e\x32\x0b.dfu.OpCode\x12\x1e\n\x04init\x18\x02 \x01(\x0b\x32\x10.dfu.InitCommand\x12 \n\x05reset\x18\x03 \x01(\x0b\x32\x11.dfu.ResetCommand"m\n\rSignedCommand\x12\x1d\n\x07\x63ommand\x18\x01 \x02(\x0b\x32\x0c.dfu.Command\x12*\n\x0esignature_type\x18\x02 \x02(\x0e\x32\x12.dfu.SignatureType\x12\x11\n\tsignature\x18\x03 \x02(\x0c"S\n\x06Packet\x12\x1d\n\x07\x63ommand\x18\x01 \x01(\x0b\x32\x0c.dfu.Command\x12*\n\x0esigned_command\x18\x02 \x01(\x0b\x32\x12.dfu.SignedCommand*\x1d\n\x06OpCode\x12\t\n\x05RESET\x10\x00\x12\x08\n\x04INIT\x10\x01*n\n\x06\x46wType\x12\x0f\n\x0b\x41PPLICATION\x10\x00\x12\x0e\n\nSOFTDEVICE\x10\x01\x12\x0e\n\nBOOTLOADER\x10\x02\x12\x19\n\x15SOFTDEVICE_BOOTLOADER\x10\x03\x12\x18\n\x14\x45XTERNAL_APPLICATION\x10\x04*D\n\x08HashType\x12\x0b\n\x07NO_HASH\x10\x00\x12\x07\n\x03\x43RC\x10\x01\x12\n\n\x06SHA128\x10\x02\x12\n\n\x06SHA256\x10\x03\x12\n\n\x06SHA512\x10\x04*t\n\x0eValidationType\x12\x11\n\rNO_VALIDATION\x10\x00\x12\x1a\n\x16VALIDATE_GENERATED_CRC\x10\x01\x12\x13\n\x0fVALIDATE_SHA256\x10\x02\x12\x1e\n\x1aVALIDATE_ECDSA_P256_SHA256\x10\x03*3\n\rSignatureType\x12\x15\n\x11\x45\x43\x44SA_P256_SHA256\x10\x00\x12\x0b\n\x07\x45\x44\x32\x35\x35\x31\x39\x10\x01'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "src.nitrokey.trussed._bootloader.nrf52_upload.dfu.dfu_cc_pb2", _globals
)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_INITCOMMAND"].fields_by_name["sd_req"]._loaded_options = None
    _globals["_INITCOMMAND"].fields_by_name["sd_req"]._serialized_options = b"\020\001"
    _globals["_OPCODE"]._serialized_start = 783
    _globals["_OPCODE"]._serialized_end = 812
    _globals["_FWTYPE"]._serialized_start = 814
    _globals["_FWTYPE"]._serialized_end = 924
    _globals["_HASHTYPE"]._serialized_start = 926
    _globals["_HASHTYPE"]._serialized_end = 994
    _globals["_VALIDATIONTYPE"]._serialized_start = 996
    _globals["_VALIDATIONTYPE"]._serialized_end = 1112
    _globals["_SIGNATURETYPE"]._serialized_start = 1114
    _globals["_SIGNATURETYPE"]._serialized_end = 1165
    _globals["_HASH"]._serialized_start = 71
    _globals["_HASH"]._serialized_end = 125
    _globals["_BOOTVALIDATION"]._serialized_start = 127
    _globals["_BOOTVALIDATION"]._serialized_end = 194
    _globals["_INITCOMMAND"]._serialized_start = 197
    _globals["_INITCOMMAND"]._serialized_end = 445
    _globals["_RESETCOMMAND"]._serialized_start = 447
    _globals["_RESETCOMMAND"]._serialized_end = 478
    _globals["_COMMAND"]._serialized_start = 480
    _globals["_COMMAND"]._serialized_end = 585
    _globals["_SIGNEDCOMMAND"]._serialized_start = 587
    _globals["_SIGNEDCOMMAND"]._serialized_end = 696
    _globals["_PACKET"]._serialized_start = 698
    _globals["_PACKET"]._serialized_end = 781
# @@protoc_insertion_point(module_scope)
